#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/pgmspace.h>
#include <stdlib.h>

volatile signed long pq=0;			//Mittlere Wirkleistung
volatile unsigned char overflow=0;	//Statusflags

volatile unsigned char DDRAM[3];	//Displaypuffer fr LED Anzeige
const unsigned char PROGMEM CGROM[]={63, 6, 91, 79, 102, 109, 125, 7, 127,111,};	//7 Segment Code

int main(void)
{ 	unsigned char stellen=0;		//Hysterese fr Stellenumschaltung
	signed int peff=0;				//Wirkleistung in 0,1W Schritten (max. ca. +/-3kW)
    ADMUX=(1<<REFS0)|(1<<REFS1);	//interne 1,2V Referenz
    ADCSRA=128+64+16+8+5;			//250kHz ADC Takt -> 2x9,6kHz Samplerate

	TCCR0A=(1<<WGM01);				//T0: CTC
	TCCR0B=5;						//7,8125kHz Timer Takt
	OCR0A=38;						//200Hz Multiplex -> 66,7Hz Frequenz
	TIMSK0=(1<<OCIE0A);				//Timer Compare Interrupt an

	DDRB=255;						//Output Ports fr LED Segmente
	DDRD=255;						//Output Ports fr LED Mux
	DIDR0=255;						//Disable PortC Inputs

    sei();

	DDRAM[0]=255;
	DDRAM[1]=255;
	DDRAM[2]=255;

	for(;;)
	{
	if (overflow&3)					//Strom/Spannung Bereichsberschreitung ?
	   {overflow=overflow&~7;
			DDRAM[0]=64;			// --- anzeigen
			DDRAM[1]=64;
			DDRAM[2]=64;}
	if (overflow&4)					//5Hz Aktualisierung ?
	   {overflow=overflow&~4;
	    peff=(pq/256)/64;			//Wirkleistung in Watt aus Mittelwert berechnen
	    if ((signed)peff<0)			//negative Leistung ?
	       {peff=abs(peff);
			DDRAM[0]=64;			// -xx anzeigen
			if (peff>=1000)			//negative Leistung grer -99W ?
				{DDRAM[1]=1+8+64;	//Fehler anzeigen (drei Striche)
				DDRAM[2]=1+8+64;}
			else
				{DDRAM[1]=pgm_read_byte(&CGROM[peff/100%10]);	//negative Leistung anzeigen
				DDRAM[2]=pgm_read_byte(&CGROM[peff/10%10]);}}
	    else
			if (peff>=10000)		//Leistung grer 1kW ?
				{DDRAM[0]=1+8+64;	//Fehler anzeigen (drei Striche)
				DDRAM[1]=1+8+64;
				DDRAM[2]=1+8+64;}
			else
			if (stellen==0)			//2 stellige Leistung mit 1 Kommazahl anzeigen
				{if (peff>=1000)
					stellen=1;
				DDRAM[0]=pgm_read_byte(&CGROM[peff/100%10]);
				DDRAM[1]=pgm_read_byte(&CGROM[peff/10%10])+128;
				DDRAM[2]=pgm_read_byte(&CGROM[peff/1%10]);}
			else					//3 stellige Leistung anzeigen
				{if (peff<800)
					stellen=0;
				DDRAM[0]=pgm_read_byte(&CGROM[peff/1000%10]);
				DDRAM[1]=pgm_read_byte(&CGROM[peff/100%10]);
				DDRAM[2]=pgm_read_byte(&CGROM[peff/10%10]);}
		}
	}
}

SIGNAL(SIG_ADC)
{
static signed int iq,uq;
static unsigned char state;
static unsigned int senden;
if (state==0) 					//Spannung gemessen
   {iq=(signed)(ADC-512);
    if ((ADC==0)||(ADC==1023))	//Spannung berlauf ?
		{overflow|=2;
      senden=0;}
    ADMUX=(1<<REFS0)|(1<<REFS1)|1;
    ADCSRA=128+64+16+8+6;
    senden++;
    if (senden>=1924)   		//5 Werte pro Sekunde
       {overflow|=4;
      senden=0;
}
    state=1;}
else 							//Spannung gemessen
   {uq=(signed)(ADC-512);
    if ((ADC==0)||(ADC==1023))	//Strom berlauf ?
		{overflow|=1;
	      senden=0;}
    ADMUX=(1<<REFS0)|(1<<REFS1)|0;
    ADCSRA=128+64+16+8+6;
    state=0;
    pq=(signed long) pq-pq/2048+(signed long)uq*iq;}	//Leistung berechnen
}

INTERRUPT(SIG_OUTPUT_COMPARE0A)	//LED Multiplex
{
static unsigned char stelle;
stelle++;
if (stelle==3)
	stelle=0;
PORTB=255;
PORTD=~DDRAM[stelle];
PORTB=~(1<<stelle);
}

